/*************************************************************************/
/* pop3s.c - pop3s module for peephole 					 */
/*  v. 0.0.0 								 */
/*  Copyright (C) 2004 Paolo Ardoino, Alessandro Di Marco		 */
/*  by Paolo Ardoino  <ardoino.gnu@disi.unige.it> 			 */
/*  									 */
/* This program is free software; you can redistribute it and/or 	 */
/* modify it under the terms of the GNU General Public License as 	 */
/* published by the Free Software Foundation; either version 2 of the 	 */
/* License, or (at your option) any later version.			 */
/* 									 */
/* This program is distributed in the hope that it will be useful, but 	 */
/* WITHOUT ANY WARRANTY; without even the implied warranty of 		 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 	 */
/* General Public License for more details.				 */
/*  									 */
/* You should have received a copy of the GNU General Public License 	 */
/* along with this program; if not, write to the Free Software 		 */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-	 */
/* 1307, USA.								 */
/*************************************************************************/

/*
 *  $Header: /var/cvsroot/peephole/src/providers/pop3s/pop3s.c,v 1.12 2004/10/28 09:28:09 dmr Exp $
 */

/*
 *  TODO:
 *
 *  -  Is OpenSSL thread safe? I hope so, but if it weren't remember to take a
 *     look at http://josefsson.org/libgsasl (GNU SASL). [dmr]
 */

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <openssl/crypto.h>
#include <errno.h>

#include <openssl/ssl.h>

#include "../providers.h"

#include "pop3s.h"

#include "../../debug.h"

static const char cvsid[]="$Id: pop3s.c,v 1.12 2004/10/28 09:28:09 dmr Exp $";

static pthread_mutex_t *lock_cs;
static long *lock_count;

static pthread_mutex_t locker = PTHREAD_MUTEX_INITIALIZER;
static int mtctr;

static unsigned long id_function(void)
{
	return ((unsigned long) pthread_self());
}

static void locking_function(int mode, int type, const char *file, int line)
{
	if (mode & CRYPTO_LOCK) {
		pthread_mutex_lock(&lock_cs[type]);
		lock_count[type]++;
	} else {
		pthread_mutex_unlock(&lock_cs[type]);
	}
}

static void ssl_thread_setup(void)
{
	int num = CRYPTO_num_locks();
	int ctr;

	lock_cs = (pthread_mutex_t*) OPENSSL_malloc(num * sizeof(pthread_mutex_t));
	lock_count = (long*) OPENSSL_malloc(num * sizeof(long));

	for (ctr = 0; ctr < num; ctr++) {
		lock_count[ctr] = 0;
		pthread_mutex_init(&lock_cs[ctr], NULL);
	}

	CRYPTO_set_id_callback(id_function);
	CRYPTO_set_locking_callback(locking_function);
}


static void ssl_thread_cleanup(void)
{
	int ctr;

	CRYPTO_set_locking_callback(NULL);

	for (ctr = 0; ctr < CRYPTO_num_locks(); ctr++) {
		pthread_mutex_destroy(&lock_cs[ctr]);
	}

	OPENSSL_free(lock_cs);
	OPENSSL_free(lock_count);
}

static void *open_pop3s(struct hostent *hent, int port, char *user, char *password)
{
	struct sockaddr_in sa;

	int sd;
	char cmd[BUFFER_SIZE];

	SSL *ssl;
	SSL_CTX *ctx;

	pthread_mutex_lock(&locker);

	if (!mtctr++) {
		ssl_thread_setup();
	}

	pthread_mutex_unlock(&locker);

	sa.sin_family = hent->h_addrtype;
	sa.sin_port = htons(port);

	memcpy((char *) &sa.sin_addr.s_addr, hent->h_addr_list[0], hent->h_length);

	if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		PRINTD("cannot open socket for '%s'\n", hent->h_name);
		return NULL;
	}

	PRINTD("connecting to '%s'\n", hent->h_name);

	if (connect(sd, (struct sockaddr *) &sa, sizeof(struct sockaddr_in)) == -1) {
		PRINTD("connection failed on '%s'\n", hent->h_name);
		close(sd);
		return NULL;
	}

	PRINTD("'%s' connected\n", hent->h_name);

	SSLeay_add_ssl_algorithms();
	ctx = SSL_CTX_new(SSLv23_client_method());
	ssl = SSL_new(ctx);
	SSL_CTX_free(ctx);
	SSL_set_fd(ssl,sd);

	if (SSL_connect(ssl) == 1) {
		READ_BUFFER(ssl, cmd);
		if (OK_BUFFER(cmd)) {
			sprintf(cmd, "USER %s\r\n", user);
			WRITE_BUFFER(ssl, cmd);
			READ_BUFFER(ssl, cmd);
			if (OK_BUFFER(cmd)) {
				sprintf(cmd, "PASS %s\r\n", password);
				WRITE_BUFFER(ssl, cmd);
				READ_BUFFER(ssl,cmd);
				if (OK_BUFFER(cmd)) {
					return ssl;
				}
			}
		}
	}

	PRINTD("cannot authenticate on '%s'\n", hent->h_name);

	if (!SSL_shutdown(ssl)) {
		SSL_shutdown(ssl);
	}

	close(sd);
	
	SSL_free(ssl);

	pthread_mutex_lock(&locker);

	if (!--mtctr) {
		ssl_thread_cleanup();
	}

	pthread_mutex_unlock(&locker);

	return NULL;
}

static void close_pop3s(void *ssl)
{
	char cmd[BUFFER_SIZE];

	sprintf(cmd, "QUIT\r\n");
	WRITE_BUFFER(ssl,cmd);
	READ_BUFFER(ssl,cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("pop3s refuses to quit\n");
	}

	if (!SSL_shutdown(ssl)) {
		SSL_shutdown(ssl);
	}

	close(SSL_get_fd(ssl));
	
	SSL_free(ssl);

	pthread_mutex_lock(&locker);

	if (!--mtctr) {
		ssl_thread_cleanup();
	}

	pthread_mutex_unlock(&locker);
}

static int stat_pop3s(void *ssl, int *msgs, int *last)
{
	char cmd[BUFFER_SIZE];

	sprintf(cmd, "STAT\r\n");
	WRITE_BUFFER(ssl,cmd);
	READ_BUFFER(ssl,cmd);
	if (ERR_BUFFER(cmd)) {
		PRINTD("error receiving stat\n");
		return -EFAULT;
	}

	*msgs = atoi(&cmd[4]);

	sprintf(cmd, "LAST\r\n");
	WRITE_BUFFER(ssl, cmd);
	READ_BUFFER(ssl, cmd);

	*last = OK_BUFFER(cmd) ? atoi(&cmd[4]) : 0;

	return 0;
}

static int match_pop3s(void *ssl, int msg, signaler_cb signaler, void *arg)
{
	char cmd[50 * BUFFER_SIZE], *loc;
	int size;

	sprintf(cmd, "TOP %d 0\r\n", msg);
	WRITE_BUFFER(ssl, cmd);

	size = SSL_read((SSL *) ssl, cmd, 50 * BUFFER_SIZE);
	if (size > 0) {
		if (ERR_BUFFER(cmd)) {
			PRINTD("message %d not found\n", msg);
			return -EFAULT;
		}

		cmd[size] = '\0';

		loc = 1 + strstr(cmd, "\n");

		signaler(loc, arg);

		return 0;
	}

	PRINTD("troubles in message %d reception\n", msg);
	return -EFAULT;
}

struct wm_providers_stub hooks = {
	open_pop3s,
	close_pop3s,
	stat_pop3s,
	match_pop3s
};
